/*
 * Filesystem Checksum Tool (fschksum) - main program
 *
 * Copyright 2021,2022 Christian Schneider <software(at)chschneider(dot)eu>
 *
 * This file is part of the Filesystem Checksum Tool (fschksum)
 * (called "program" in the following).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation, *not* any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <iomanip>
#include <filesystem>
#include <ctime>
#include <string>
#include <fstream>
#include <array>
#include <functional>
#include <vector>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <cppdb/frontend.h>
#include <openssl/md5.h>
#include "version.hpp"



namespace fs = std::filesystem;

using md5_t   = std::array<unsigned char, 16>;
using stat_t  = struct stat;
using sqlid_t = long long int;

constexpr size_t file_buf_size = 1024 * 1024;

class myexception : public std::exception {
  public:
    myexception(const std::string& msg = "myexception exception")
      throw(): msg_(msg) {}
    virtual ~myexception() throw() {}
    virtual const char* what() const throw() { return msg_.c_str(); };
  private:
    std::string msg_;
};



// create tables
inline std::string create_tables(cppdb::session& sql,
                                 const std::string basedir) {
  cppdb::statement stmt_idx = sql << "INSERT INTO toc (basedir) VALUES (?);"
                                  << basedir << cppdb::exec;
  std::string id_table = std::to_string(stmt_idx.last_insert_id());
  sql << "CREATE TABLE inodes" + id_table + "("
         "  inode      INTEGER PRIMARY KEY NOT NULL,"
         "  mtime      INTEGER,"
         "  size       INTEGER,"
         "  md5sum     TEXT,"
         "  type       INTEGER"
         ");" << cppdb::exec;
  sql << "CREATE TABLE tree" + id_table + " ("
         "  id         INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,"
         "  id_parent  INTEGER,"
         "  inode      INTEGER,"
         "  name       TEXT,"
         "  CONSTRAINT cn_tree" + id_table + "_parent FOREIGN KEY (id_parent)"
         "    REFERENCES tree" + id_table + "(id)"
         "  CONSTRAINT cn_tree" + id_table + "_inode  FOREIGN KEY (inode)"
         "    REFERENCES inodes" + id_table + "(inode)"
         ");" << cppdb::exec;
  return id_table;
}

// recursively build the full paths for regular files and select one per inode
inline std::string fullpath_cte(const std::string& id_table,
                                const std::string& db_prefix = "") {
  return
    "fullpath_cte (id, id_parent, inode, name, level, fullpath) AS"
    "( SELECT id, id_parent, inode, name, 0 AS level,"
    "    CAST(name AS TEXT) AS fullpath"
    "  FROM " + db_prefix + "tree" + id_table +
    "  WHERE id_parent IS NULL"
    "  UNION ALL"
    "  SELECT t.id AS id, t.id_parent, t.inode, t.name,"
    "    fullpath_cte.level + 1 AS level,"
    "    CAST(fullpath_cte.fullpath || '/' || CAST(t.name AS TEXT) AS TEXT)"
    "    AS fullpath"
    "  FROM " + db_prefix + "tree" + id_table + " t"
    "  INNER JOIN fullpath_cte"
    "  ON fullpath_cte.id = t.id_parent )";
}

// find *all* descendants recursively of a node
inline std::string descendants_cte(const std::string& id_table) {
  return
    "descendants_cte (id_parent, id_child, inode_child, level, name) AS"
    "( SELECT id_parent, id, inode, 1, name FROM tree" + id_table +
    "  UNION ALL"
    "  SELECT c.id_parent, t.id, t.inode, c.level + 1,"
    "    CAST(c.name || '/' || t.name AS TEXT) fullname"
    "  FROM descendantscte c"
    "  JOIN tree" + id_table + " t"
    "  ON c.id_child = t.id_parent )";
}



void do_stat(const fs::path& p, stat_t* const s) {
  if (s == nullptr)
    throw myexception("do_stat(): no output variable `s' supplied.");

  if (stat(p.c_str(), s))
    throw myexception("do_stat(): problem retrieving file status for `"
                      + p.string() + "'.");
}

void calc_md5_file(const fs::path& p, md5_t* const md5) {
  if (md5 == nullptr)
    throw myexception("calc_md5_file(): no output variable `md5' supplied.");

  if (! fs::is_regular_file(p))
    throw myexception("calc_md5_file(): file `" + p.string()
                      + "' is not a regular file.");

  std::ifstream ifn(p);
  if (! ifn)
    throw myexception("calc_md5_file(): problem opening file `"
                      + p.string() + "'.");

  MD5_CTX c;
  if (! MD5_Init(&c))
    throw myexception("calc_md5_file(): problem initializing MD5 routine.");

  std::array<char, file_buf_size> buf;
  while (ifn.read(buf.data(), buf.size())) {
    if (! MD5_Update(&c, buf.data(), buf.size()))
      throw myexception("calc_md5_file(): problem updating MD5 routine.");
  }
  if (! MD5_Update(&c, buf.data(), ifn.gcount() % buf.size()))
    throw myexception("calc_md5_file(): problem updating MD5 routine at end.");

  if (! MD5_Final(md5->data(), &c))
    throw myexception("calc_md5_file(): problem finalizing MD5 routine.");
}

void calc_md5_string(const std::string& s, md5_t* const md5) {
  if (md5 == nullptr)
    throw myexception("calc_md5_file(): no output variable `md5' supplied.");

  MD5_CTX c;
  if (! MD5_Init(&c))
    throw myexception("calc_md5_file(): problem initializing MD5 routine.");
  if (! MD5_Update(&c, s.data(), s.size()))
    throw myexception("calc_md5_file(): problem updating MD5 routine.");
  if (! MD5_Final(md5->data(), &c))
    throw myexception("calc_md5_file(): problem finalizing MD5 routine.");
}

// this is a quick hack
inline std::string md5asc(const md5_t& md5) {
  std::string s;
  for (auto c: md5) {
    for (auto x: { (c & 0xf0) >> 4, (c & 0x0f) }) {
      if (x <= 9)
        s = s + static_cast<char>('0' + x);
      else
        s = s + static_cast<char>('a' + x - 10);
    }
  }
  return s;
}

void do_find(const fs::path& parent,
             cppdb::statement& stmt_inode,
             cppdb::statement& stmt_tree,
             const sqlid_t id_parent = -1) {
  for (auto& p: fs::directory_iterator(parent)) {
    stat_t s;
    if (p.is_symlink()) {
      // ignore; this explicit check is required, as is_directory() and
      // is_regular_file(), respectively, seem to be true for symlinks!?
    }
    else if (p.is_directory()) {
      do_stat(p, &s);
      stmt_inode.reset();
      // set mtime and size for directories later when hashing directories
      stmt_inode << s.st_ino << cppdb::null << cppdb::null << 0 << cppdb::exec;
      stmt_tree.reset();
      if (id_parent == -1)
        stmt_tree << cppdb::null << s.st_ino << p.path().filename()
                  << cppdb::exec;      
      else
        stmt_tree << id_parent << s.st_ino << p.path().filename()
                  << cppdb::exec;
      do_find(p.path(), stmt_inode, stmt_tree, stmt_tree.last_insert_id());
    }
    else if (p.is_regular_file()) {
      do_stat(p, &s);
      stmt_inode.reset();
      stmt_inode << s.st_ino << s.st_mtim.tv_sec << s.st_size << 1
                 << cppdb::exec;
      stmt_tree.reset();
      stmt_tree << id_parent << s.st_ino << p.path().filename() << cppdb::exec;
    }
    else {
      // ignore anything else
    }
  }
}

class importer_ifs {
  public:
    struct importer_t {
      ino_t       inode;
      time_t      mtime;
      off_t       size;
      char        type;
      std::string fullpath;
      std::string path;
      std::string file;
    };

  public:
    importer_ifs(const std::string fn) : ifs_(fn) {
      if (ifs_)
        read();
      else
        throw myexception("in import command: problem opening FILES file.");
    }
    void read() {
      if (ifs_ >> curr_.inode >> curr_.mtime >> curr_.size >> curr_.type) {
        std::string str;
        if (std::getline(ifs_, str)) {
          curr_.fullpath = str.substr(1);
          std::size_t found = str.find_last_of("/");
          if (found == std::string::npos) {
            curr_.path = "";
            curr_.file = str.substr(1);
          }
          else {
            curr_.path = str.substr(1, found - 1);
            curr_.file = str.substr(found + 1);            
          }
        }
      }
    }
    operator bool() const {
      return ifs_ ? true : false;
    }
    const importer_t& curr() const {
      return curr_;
    }
  private:
    std::ifstream ifs_;
    importer_t    curr_;
};

void do_import(importer_ifs& ifp,
               cppdb::statement& stmt_inode, cppdb::statement& stmt_tree,
               const std::string& path, const sqlid_t id_parent = -1) {
  while (ifp && path == ifp.curr().path) {
    if (ifp.curr().type == 'd') {
      stmt_inode.reset();
      stmt_inode << ifp.curr().inode << cppdb::null << cppdb::null << 0
                 << cppdb::exec;
      stmt_tree.reset();
      if (id_parent == -1)
        stmt_tree << cppdb::null << ifp.curr().inode << ifp.curr().file
                  << cppdb::exec;
      else
        stmt_tree << id_parent << ifp.curr().inode << ifp.curr().file
                  << cppdb::exec;
      const std::string path = ifp.curr().fullpath;
      ifp.read();
      do_import(ifp, stmt_inode, stmt_tree,
                path, stmt_tree.last_insert_id());
    }
    else if (ifp.curr().type == 'f') {
      stmt_inode.reset();
      stmt_inode << ifp.curr().inode << ifp.curr().mtime
                 << ifp.curr().size << 1 << cppdb::exec;
      stmt_tree.reset();
      if (id_parent == -1)
        stmt_tree << cppdb::null << ifp.curr().inode << ifp.curr().file
                  << cppdb::exec;
      else
        stmt_tree << id_parent << ifp.curr().inode << ifp.curr().file
                  << cppdb::exec;
      ifp.read();
    }
    else {
      // ignore anything else
    }
  }
}

// calculate for *directories*:
// * size = size of all descendants,
// * mtime = largest mtime of any descendant, and
// * MD5 checksum = MD5 sum of concatenated MD5 sums of direct children
// NOTE: In order to obtain deterministic MD5 sum results for a directory
//   structure (= *independent* of the inode numbering etc., hence, the order
//   in which the tree is generated above), we rely on two facts:
// (1) a tree `id' is always greater than its `id_parent', hence, sorting
//   `parent.id' below in *descending* order will always return all the
//   descendants before any ancestor tree node; when calculating MD5 sums,
//   this ensures that all the descentants' MD5 sums are available, when an
//   ancestor node's MD5 sum is calculated;
// (2) direct children nodes are sorted according to their names resulting
//   in a deterministic order in the concatenation of the MD5 sums.
void hash_dirs(cppdb::session& sql, const std::string& id_table) {
  // with some other optimizations and indexing introduced in the main code
  // below, the temp table might not provide any advantage anymore, but is left
  // for now to not break things:
  sql << "CREATE TEMP TABLE children_tt AS"
         "  SELECT parent.id AS id_parent,"
         "    parent.inode AS inode_parent, parent.name AS name_parent,"
         "    child.inode AS inode_child, child.name AS name_child"
         "  FROM tree" + id_table + " parent"
         "  JOIN tree" + id_table + " child"
         "  ON parent.id = child.id_parent"
         "  ORDER BY parent.id ASC, child.name ASC;"
      << cppdb::exec;
  sql << "CREATE INDEX TEMP.cp_idx ON children_tt(id_parent);"
      << cppdb::exec;
  sql << "CREATE INDEX TEMP.cc_idx ON children_tt(inode_child);"
      << cppdb::exec;

  cppdb::result rslt_dirs = sql <<
    "SELECT id_parent, inode_parent"
    "  FROM TEMP.children_tt"
    "  GROUP BY id_parent"
    "  ORDER BY id_parent DESC;";
  cppdb::statement stmt_children = sql <<
    "SELECT MAX(i.mtime), SUM(i.size), GROUP_CONCAT(i.md5sum, \"\")"
    "  FROM TEMP.children_tt c"
    "  JOIN inodes" + id_table + " i"
    "  ON c.inode_child = i.inode"
    "  WHERE c.id_parent = ?;";
  cppdb::statement stmt_inode_upd = sql <<
    "UPDATE inodes" + id_table +
    "  SET mtime = ?, size = ?, md5sum = ?"
    "  WHERE inode = ?;";

  while (rslt_dirs.next()) {
    sqlid_t id_dir;
    ino_t   inode_dir;
    rslt_dirs >> id_dir >> inode_dir;

    stmt_children.reset();
    stmt_children.bind(id_dir);
    cppdb::result rslt_children = stmt_children.query();
    time_t mtime = 0;
    off_t  size  = 0;
    std::string md5str;

    if (rslt_children.next() && ! rslt_children.is_null(0))
      rslt_children >> mtime >> size >> md5str;
    if (rslt_children.next())
      throw myexception("Second result during hashing of directory ID `"
                        + std::to_string(id_dir) + "'.");

    md5_t md5;
    calc_md5_string(md5str, &md5);
    stmt_inode_upd.reset();
    stmt_inode_upd << mtime << size << md5asc(md5) << inode_dir
                   << cppdb::exec;
    //std::cerr << id_dir << "\t" << md5asc(md5) << "\t" << md5str << std::endl;
  }
}



int main(int argc, char* argv[]) {
  if (argc < 3 || argc > 4) {
    std::cerr << "usage: " << argv[0] << " <command> <database> [ <extra arg> ]" << std::endl
              << std::endl
              << "available commands:" << std::endl
              << "  create         create database entries; calculate MD5 checksums of all files;"  << std::endl
              << "                 the additional `<extra arg>' must specify the base directory"    << std::endl
              << "  update         update an existing database; do *not* recalculate an MD5"        << std::endl
              << "                 checksum, if both size and mtime of a file are unchanged"        << std::endl
              << "  import         import files generated with the shell script (if information"    << std::endl
              << "                 needs to be gathered on a host with minimal or outdated system"  << std::endl
              << "                 not supporting this tool); `<extra arg>' must be stem of the"    << std::endl
              << "                  `*.FILES' and `*.MD5SUMS' files"                                << std::endl
              << "  check          check filesystem against database (all hardlinks, if any)"       << std::endl
              << "  check-sparse   check filesystem against database (only one of each hardlink)"   << std::endl
              << "  export         write `MD5SUMS' format of database to stdout (all hardlinks, if" << std::endl
              << "                 any)"                                                            << std::endl
              << "  export-sparse  write `MD5SUMS' format of database to stdout (only one of each"  << std::endl
              << "                 hardlink)"                                                       << std::endl
              << "  diff           compare two database updates; if an `<extra arg>' is given, it"  << std::endl
              << "                 must refer to another database and the latest update of each"    << std::endl
              << "                 database are compared; otherwise, the latested two updates"      << std::endl
              << "                 within `<database>' are compared"                                << std::endl
              << "  diff-verbose   like `diff', but also prints size, mtime, and MD5 checksums"     << std::endl
              << "  diff-quick     similar to `diff', but considers directories identical, if both" << std::endl
              << "                 their size and MD5 checksum match; good to check for identical"  << std::endl
              << "                 content in a file/directory name agnostic way (also"             << std::endl
              << "                 disregarding empty directories"                                  << std::endl
              << "  expire         remove all but the latest update"                                << std::endl
              << "  stats          print statistics"                                                << std::endl
              << std::endl
              << "revision:  " << revision << std::endl
              << "copyright: 2021,2022 Christian Schneider <software(at)chschneider(dot)eu>" << std::endl
              << "license:   GNU GPL v3 (not any other version!); USE ON YOUR OWN RISK!" << std::endl;
    return 1;
  }

  // very simple command-line parser
  std::string cmd_str = argv[1];
  if (   cmd_str != "create"
      && cmd_str != "update"
      && cmd_str != "check"
      && cmd_str != "check-sparse"
      && cmd_str != "export"
      && cmd_str != "export-sparse"
      && cmd_str != "import"
      && cmd_str != "diff"
      && cmd_str != "diff-verbose"
      && cmd_str != "diff-quick"
      && cmd_str != "expire"
      && cmd_str != "stats") {
    std::cerr << "ERROR: Unsupported command `" << cmd_str << "'." << std::endl;
    return 1;
  }

  std::string db_name = argv[2];
  std::string db_mode;
  std::string basedir = (argc == 4) ? argv[3] : "";
  // strip all trailing "/", if any
  while (   basedir.size() > 1
         && basedir[basedir.size() - 1] == '/')
    basedir = basedir.substr(0, basedir.size() - 1);

  // check for existence of database
  if (cmd_str == "create" || cmd_str == "import") {
    if (fs::exists(db_name)) {
      std::cerr << "ERROR: database `" << db_name << "' exist already."
                << std::endl;
      return 1;
    }
  }
  else {
    if (! fs::exists(db_name)) {
      std::cerr << "ERROR: database `" << db_name << "' does not exist."
                << std::endl;
      return 1;
    }
    if (! fs::is_regular_file(db_name)) {
      std::cerr << "ERROR: database `" << db_name << "' is not a regular file."
                << std::endl;
      return 1;
    }
  }

  // open database read-only, if possible
  if (   cmd_str == "check"  || cmd_str == "check-sparse"
      || cmd_str == "export" || cmd_str == "export-sparse"
      || cmd_str == "stats"  || cmd_str == "diff-quick"
      || cmd_str == "diff"   || cmd_str == "diff-verbose") {
    db_mode = ";mode=readonly";
  }

  // check second argument
  if (cmd_str == "create") {
    if (basedir.empty()) {
      std::cerr << "ERROR: no `<basedir>' specified." << std::endl;
      return 1;
    }
    if (! fs::is_directory(basedir)) {
      std::cerr << "ERROR: `" << basedir << "' is not a directory."
                << std::endl;
      return 1;
    }
  }
  else if (cmd_str == "import") {
    if (basedir.empty()) {
      std::cerr << "ERROR: no `<basedir>' file stem specified." << std::endl;
      return 1;
    }
    if (! fs::exists(basedir + ".FILES")) {
      std::cerr << "ERROR: `" << basedir << ".FILES' does not exist."
                << std::endl;
      return 1;
    }
    if (! fs::is_regular_file(basedir + ".FILES")) {
      std::cerr << "ERROR: `" << basedir << ".FILES' is not a regular file."
                << std::endl;
      return 1;
    }
    if (! fs::exists(basedir + ".MD5SUMS")) {
      std::cerr << "ERROR: `" << basedir << ".MD5SUMS' does not exist."
                << std::endl;
      return 1;
    }
    if (! fs::is_regular_file(basedir + ".MD5SUMS")) {
      std::cerr << "ERROR: `" << basedir << ".MD5SUMS' is not a regular file."
                << std::endl;
      return 1;
    }
  }
  else if (   cmd_str == "diff"
           || cmd_str == "diff-verbose"
           || cmd_str == "diff-quick") {
    if (! basedir.empty()) {
      if (! fs::exists(basedir)) {
        std::cerr << "ERROR: `<other database>' does not exist." << std::endl;
        return 1;
      }
      if (! fs::is_regular_file(basedir)) {
        std::cerr << "ERROR: `<other database>' is not a regular file."
                  << std::endl;
        return 1;
      }
    }
  }
  else {
    if (! basedir.empty()) {
      std::cerr << "ERROR: `<basedir>' must not be specified." << std::endl;
      return 1;
    }
  }



  int retval = 0;
  try {
    cppdb::session sql("sqlite3:db=" + db_name + db_mode);
    sql << "PRAGMA foreign_keys = ON;" << cppdb::exec;
    sql << "CREATE TABLE IF NOT EXISTS toc ("
            "  id       INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,"
            "  created  DATETIME DEFAULT CURRENT_TIMESTAMP,"
            "  basedir  TEXT NOT NULL"
            ");" << cppdb::exec;

    // obtain last database tables from "toc" table, if any
    std::string id_last, created;
    if (   cmd_str != "diff"
        && cmd_str != "diff-verbose"
        && cmd_str != "diff-quick") {
      if (basedir == "") {
        cppdb::result r = sql << "SELECT MAX(id), basedir, created FROM toc;"
                              << cppdb::row;
        if (! r.empty() && ! r.is_null(0))
          r >> id_last >> basedir >> created;
        else
          throw myexception("No previous database entries found.");
      }
      else {
        cppdb::result r = sql << "SELECT MAX(id), created FROM toc"
                                 "  WHERE basedir = ?;"
                              << basedir << cppdb::row;
        if (! r.empty() && ! r.is_null(0))
          r >> id_last >> created;
      }
    }



    if (cmd_str == "create" || cmd_str == "update") {
      cppdb::transaction guard(sql);
      const std::string id_table = create_tables(sql, basedir); 

      // find all directories and regular files and record their inodes
      std::cerr << "searching filesystem tree ..." << std::endl;
      cppdb::statement stmt_inode = sql <<
        "INSERT OR IGNORE INTO inodes" + id_table +
        "  (inode, mtime, size, type)"
        "  VALUES (?, ?, ?, ?)";
      cppdb::statement stmt_tree = sql <<
        "INSERT INTO tree" + id_table +
        "  (id_parent, inode, name)"
        "  VALUES (?, ?, ?)";
      stat_t s;
      do_stat(basedir, &s);
      stmt_inode << s.st_ino << cppdb::null << cppdb::null << 0 << cppdb::exec;
      stmt_tree << cppdb::null << s.st_ino << "." << cppdb::exec;

      do_find(basedir, stmt_inode, stmt_tree, stmt_tree.last_insert_id());
      stmt_inode.clear();
      stmt_tree.clear();

      // create indexes to speed stuff up
      std::cerr << "creating index on tree (id_parent) ..." << std::endl;
      sql << "CREATE INDEX idx_id_parent_tree" + id_table +
             "  ON tree" + id_table + " (id_parent);"
          << cppdb::exec;
      std::cerr << "creating index on tree (inode) ..." << std::endl;
      sql << "CREATE INDEX idx_inode_tree" + id_table +
             "  ON tree" + id_table + " (inode);"
          << cppdb::exec;

      // if updating, transfer unchanged MD5SUMS over first
      cppdb::statement stmt_md5 = sql <<
        "UPDATE inodes" + id_table + " SET md5sum = ?"
        "  WHERE inode = ?";
      if (cmd_str == "update" && ! id_last.empty()) {
        std::cerr << "transferring MD5 checksums ..."
                  << std::endl;
        cppdb::result rslt_md5 = sql <<
          "SELECT i1.inode, i1.md5sum FROM inodes" + id_last + " i1"
          "  JOIN inodes" + id_table + " i2"
          "  ON i1.inode = i2.inode"
          "  WHERE i1.mtime = i2.mtime"
          "    AND i1.size = i2.size"
//          "  WHERE i1.size = i2.size"
          "    AND i1.type = i2.type;";
        while (rslt_md5.next()) {
          ino_t inode = 0;  
          std::string md5;
          rslt_md5 >> inode >> md5;
          stmt_md5.reset();
          stmt_md5 << md5 << inode << cppdb::exec;
        }
      }

      // create MD5 checksums for each *regular* *file*
      std::cerr << "calculating MD5 checksums ..."
                << std::endl;
      //cppdb::result rslt_ino = sql << "WITH " + fullpath_cte(id_table) +
      //  "SELECT f.inode, MIN(f.fullpath) FROM fullpath_cte f"
      //  "  JOIN inodes" + id_table + " i"
      //  "  ON i.inode = f.inode"
      //  "  WHERE i.md5sum IS NULL AND i.type = 1"
      //  "  GROUP BY i.inode;";
      cppdb::result rslt_ino = sql <<
        "WITH null_md5_cte (id) AS"
        "  ( SELECT t.id"
        "    FROM tree" + id_table + " t"
        "    JOIN inodes" + id_table + " i"
        "    ON t.inode = i.inode"
        "    WHERE i.md5sum IS NULL AND i.type = 1"
        "    GROUP BY t.inode ),"
        "filepath_cte (id, id_parent, inode, level, fullpath) AS"
        "  ( SELECT id, id_parent, inode, 0 AS level, name AS fullpath"
        "    FROM tree" + id_table +
        "    WHERE id IN (SELECT * FROM null_md5_cte)"
        "    UNION ALL"
        "    SELECT t.id, t.id_parent, f.inode, f.level + 1 AS level,"
        "      CAST(t.name || '/' || f.fullpath AS TEXT) AS fullpath"
        "    FROM tree" + id_table + " t"
        "    INNER JOIN filepath_cte f"
        "    ON f.id_parent = t.id )"
        "SELECT inode, fullpath FROM filepath_cte WHERE id_parent IS NULL;";

      while (rslt_ino.next()) {
        ino_t inode = 0;  
        std::string p;
        rslt_ino >> inode >> p;
        p = basedir + "/" + p;

        md5_t md5;
        if (fs::is_regular_file(p)) {
          std::cout << "  " << p << std::endl;
          calc_md5_file(p, &md5);
          stmt_md5.reset();
          stmt_md5 << md5asc(md5) << inode << cppdb::exec;
        }
      }

      std::cerr << "hashing directories ..." << std::endl;
      hash_dirs(sql, id_table);

      guard.commit();
    }



    if (cmd_str == "import") {
      cppdb::transaction guard(sql);

      const std::string id_table = create_tables(sql, basedir);

      cppdb::statement stmt_inode = sql <<
        "INSERT OR IGNORE INTO inodes" + id_table + " (inode, mtime, size, type)"
        "VALUES (?, ?, ?, ?)";
      cppdb::statement stmt_tree = sql <<
        "INSERT INTO tree" + id_table + " (id_parent, inode, name)"
        "VALUES (?, ?, ?)";

      importer_ifs ifp(basedir + ".FILES");
      do_import(ifp, stmt_inode, stmt_tree, ifp.curr().path);

      // create indexes to speed stuff up
      std::cerr << "creating index on tree (id_parent) ..." << std::endl;
      sql << "CREATE INDEX idx_id_parent_tree" + id_table +
             "  ON tree" + id_table + " (id_parent);"
          << cppdb::exec;
      std::cerr << "creating index on tree (inode) ..." << std::endl;
      sql << "CREATE INDEX idx_inode_tree" + id_table +
             "  ON tree" + id_table + " (inode);"
          << cppdb::exec;

      cppdb::statement stmt_inode_md5 = sql <<
        "UPDATE inodes" + id_table + " SET md5sum = ?"
        "  WHERE inode = ? AND mtime = ?;";

      std::ifstream ifs_md5(basedir + ".MD5SUMS");
      if (! ifs_md5)
        throw myexception("in import command: problem opening MD5SUMS file.");
      ino_t       inode;
      time_t      mtime;
      std::string md5sum;
      while (ifs_md5 >> inode >> mtime >> md5sum) {
        stmt_inode_md5.reset();
        stmt_inode_md5 << md5sum << inode << mtime << cppdb::exec;
      }

      hash_dirs(sql, id_table);

      guard.commit();
    }



    if (cmd_str == "check" || cmd_str == "check-sparse") {
      if (id_last.empty())
        throw myexception("No entries ");

      std::string sql_group_by = "";
      if (cmd_str == "check-sparse")
        sql_group_by = "  GROUP BY f.inode";

      // check MD5 sums
      cppdb::result rslt_chk_md5 = sql << "WITH " + fullpath_cte(id_last) +
        "SELECT f.fullpath, i.md5sum, i.mtime, i.size FROM fullpath_cte f"
        "  JOIN inodes" + id_last + " i"
        "  ON f.inode = i.inode"
        "  WHERE i.type = 1" +
        sql_group_by +
        "  ORDER BY f.fullpath;";
      while (rslt_chk_md5.next()) {
        std::string p;
        std::string md5_old;
        time_t mtime;
        off_t  size;
        rslt_chk_md5 >> p >> md5_old >> mtime >> size;
        // add absolute basedir and remove trailing "./"
        p = basedir + "/" + p.substr(2);

        md5_t md5;
        if (! fs::is_regular_file(p))
          std::cout <<   "DEL  " << p << std::endl;
        else {
          stat_t s;
          do_stat(p, &s);

          // calculate MD5 sum only, if size is the same
          bool same = (s.st_size == size);
          if (same) {
            calc_md5_file(p, &md5);
            same = (md5_old == md5asc(md5));
          }

          if (same) {
            if (s.st_mtim.tv_sec == mtime)
              std::cout << "OK   " << p << std::endl;
            else
              std::cout << "OK*  " << p << std::endl;
          }
          else {
            if (s.st_mtim.tv_sec == mtime) {
              std::cerr << "ERR  " << p << std::endl;
              retval = 2;
            }
            else
              std::cout << "MOD  " << p << std::endl;
          }
        }
      }
    }



    if (cmd_str == "export" || cmd_str == "export-sparse") {
      std::string sql_group_by = "";
      if (cmd_str == "export-sparse")
        sql_group_by = "  GROUP BY f.inode";

      cppdb::result rslt_md5s = sql << "WITH " + fullpath_cte(id_last) +
        "SELECT i.md5sum, f.fullpath"
        "  FROM inodes" + id_last + " i"
        "  JOIN fullpath_cte f"
        "  ON f.inode = i.inode"
        "  WHERE i.type = 1" +
        sql_group_by +
        "  ORDER BY f.fullpath";
      while (rslt_md5s.next()) {
        std::string md5;
        std::string name;
        rslt_md5s >> md5 >> name;
        std::cout << md5 << "  " << name << std::endl;
      }
    }



    if (   cmd_str == "diff"
        || cmd_str == "diff-verbose"
        || cmd_str == "diff-quick") {
      std::string id_lhs, id_rhs, basedir_lhs, basedir_rhs,
                  created_lhs, created_rhs, db_rhs;
      const bool verbose = (cmd_str == "diff-verbose");

      cppdb::result r = sql << "SELECT MAX(id), created, basedir FROM toc;"
                            << cppdb::row;
      if (! r.empty() && ! r.is_null(0))
        r >> id_lhs >> created_lhs >> basedir_lhs;

      // look for second-newest entries in same database
      if (basedir == "") {
        cppdb::result r = sql << "SELECT MAX(id), created, basedir FROM toc"
                                 "  WHERE id <> ?;" << id_lhs << cppdb::row;
        if (! r.empty() && ! r.is_null(0))
          r >> id_rhs >> created_rhs >> basedir_rhs;
        else
          throw myexception("No other database entries found for diff.");
        std::swap(id_rhs,      id_lhs);
        std::swap(created_rhs, created_lhs);
        std::swap(basedir_rhs, basedir_lhs);
      }
      // look for newest entry in other database
      else {
        sql << "ATTACH DATABASE \"file:" + basedir + "?mode=ro\" AS db;"
            << cppdb::exec;
        db_rhs = "db.";

        cppdb::result r = sql << "SELECT MAX(id), created, basedir FROM db.toc;"
                              << cppdb::row;
        if (! r.empty() && ! r.is_null(0))
          r >> id_rhs >> created_rhs >> basedir_rhs;
        else
          throw myexception("No other database entries found for diff.");
      }

      std::cout << "##  " << db_name << ":" << basedir_lhs << " @ " << id_lhs
                << " (" << created_lhs << ") <=> "
                << (db_rhs.empty() ? db_name : basedir) << ":" << basedir_rhs
                << " @ " << id_rhs << " (" << created_rhs << ")" << std::endl;

      std::function<void(sqlid_t, sqlid_t, const std::string&)>
      do_diff
      = [&](const sqlid_t id_lhs_parent,
            const sqlid_t id_rhs_parent,
            const std::string& path) -> void {
        std::string cmp_parent =   (id_lhs_parent == 0 && id_rhs_parent == 0)
                                 ? " IS NULL"
                                 : " = ?";

        // common portion of diff query
        std::string diff_str =
          "WITH lhs_children_cte (name, id, mtime, size, md5sum, type) AS ("
          "  SELECT lt.name, lt.id, li.mtime, li.size, li.md5sum, li.type"
          "  FROM tree" + id_lhs + " lt"
          "  JOIN inodes" + id_lhs + " li"
          "  ON lt.inode = li.inode"
          "  WHERE lt.id_parent " + cmp_parent +
          "),"
          "rhs_children_cte (name, id, mtime, size, md5sum, type) AS ("
          "  SELECT rt.name, rt.id, ri.mtime, ri.size, ri.md5sum, ri.type"
          "  FROM " + db_rhs + "tree" + id_rhs + " rt"
          "  JOIN " + db_rhs + "inodes" + id_rhs + " ri"
          "  ON rt.inode = ri.inode"
          "  WHERE rt.id_parent " + cmp_parent +
          "),"
          "diff_cte (lname, lid, lmtime, lsize, lmd5sum, ltype,"
          "          rname, rid, rmtime, rsize, rmd5sum, rtype) AS ("
          "  SELECT l.*, r.*"
          "    FROM lhs_children_cte l"
          "    LEFT JOIN rhs_children_cte r"
          "    ON l.name = r.name"
          "  UNION"
          "  SELECT l.*, r.*"
          "    FROM rhs_children_cte r"
          "    LEFT JOIN lhs_children_cte l"
          "    ON l.name = r.name"
          ")";

        // command-specific portions
        if (cmd_str == "diff")
          diff_str = diff_str +
            "SELECT lname, rname, lid, rid, ltype, rtype FROM diff_cte"
            "  WHERE lname IS NULL OR rname IS NULL OR ltype = 0 OR rtype = 0"
            "    OR (ltype = 1 AND rtype = 1"
            "      AND (rsize <> lsize OR rmd5sum <> lmd5sum))"
            "  ORDER BY COALESCE(lname, rname)";
        else if (cmd_str == "diff-verbose")
          diff_str = diff_str +
            "SELECT lname, rname, lid, rid, ltype, rtype, lmtime, rmtime,"
            "    lsize, rsize, lmd5sum, rmd5sum FROM diff_cte"
            "  WHERE lname IS NULL OR rname IS NULL OR ltype = 0 OR rtype = 0"
            "    OR (ltype = 1 AND rtype = 1"
            "      AND (rsize <> lsize OR rmd5sum <> lmd5sum))"
            "  ORDER BY COALESCE(lname, rname)";
        else if (cmd_str == "diff-quick")
          diff_str = diff_str +
          "SELECT lname, rname, lid, rid, ltype, rtype FROM diff_cte"
          "  WHERE lname IS NULL OR rname IS NULL OR rsize <> lsize"
          "        OR rmd5sum <> lmd5sum OR rtype <> ltype"
          "  ORDER BY COALESCE(lname, rname)";

        cppdb::statement stmt_diff = sql << diff_str;

        stmt_diff.reset();
        if (id_lhs_parent != 0 && id_rhs_parent != 0) {
          stmt_diff.bind(id_lhs_parent);
          stmt_diff.bind(id_rhs_parent);
        }
        cppdb::result rslt_diff = stmt_diff.query();
        while (rslt_diff.next()) {
          std::string lname, rname;
          sqlid_t lid = 0;
          sqlid_t rid = 0;
          sqlid_t ltype = -1;
          sqlid_t rtype = -1;

          const std::string xpath = (path.empty()) ? "" : path + "/";
          if (rslt_diff.is_null(0)) {
            rslt_diff.fetch(1, rname);
            rslt_diff.fetch(5, rtype);
            if (rtype == 0) {
              std::cout << "+d  ";
              if (verbose) {
                std::string mtime, size, md5sum;
                rslt_diff.fetch(7, mtime);
                rslt_diff.fetch(9, size);
                rslt_diff.fetch(11, md5sum);
                std::cout << mtime << "  " << std::setw(8) << size << "  "
                          << md5sum << "  ";
              }
              std::cout << xpath << rname << std::endl;
            }
            else if (rtype == 1) {
              std::cout << "+f  ";
              if (verbose) {
                std::string mtime, size, md5sum;
                rslt_diff.fetch(7, mtime);
                rslt_diff.fetch(9, size);
                rslt_diff.fetch(11, md5sum);
                std::cout << mtime << "  " << std::setw(8) << size << "  "
                          << md5sum << "  ";
              }
              std::cout  << xpath << rname << std::endl;
            }
            else
              throw myexception("Unknown type in database entry.");
          }
          else if (rslt_diff.is_null(1)) {
            rslt_diff.fetch(0, lname);
            rslt_diff.fetch(4, ltype);
            if (ltype == 0) {
              std::cout << "-d  ";
              if (verbose) {
                std::string mtime, size, md5sum;
                rslt_diff.fetch(6, mtime);
                rslt_diff.fetch(8, size);
                rslt_diff.fetch(10, md5sum);
                std::cout << mtime << "  " << std::setw(8) << size << "  "
                          << md5sum << "  ";
              }
              std::cout  << xpath << lname << std::endl;
            }
            else if (ltype == 1) {
              std::cout << "-f  ";
              if (verbose) {
                std::string mtime, size, md5sum;
                rslt_diff.fetch(6, mtime);
                rslt_diff.fetch(8, size);
                rslt_diff.fetch(10, md5sum);
                std::cout << mtime << "  " << std::setw(8) << size << "  "
                          << md5sum << "  ";
              }
              std::cout  << xpath << lname << std::endl;
            }
            else
              throw myexception("Unknown type in database entry.");
          }
          else {
            rslt_diff >> lname >> rname >> lid >> rid >> ltype >> rtype;
            if (ltype == 0 && rtype == 0)
              do_diff(lid, rid, xpath + lname);
            else if (ltype == 1 && rtype == 1) {
              std::cout << "*f  ";
              if (verbose) {
                std::string mtime, size, md5sum;
                rslt_diff.fetch(7, mtime);
                rslt_diff.fetch(9, size);
                rslt_diff.fetch(11, md5sum);
                std::cout << mtime << "  " << std::setw(8) << size << "  "
                          << md5sum << "  ";
              }
              std::cout  << xpath << lname << std::endl;
            }
            else if (ltype == 0 && rtype == 1) {
              std::cout << "-d  ";
              if (verbose) {
                std::string mtime, size, md5sum;
                rslt_diff.fetch(6, mtime);
                rslt_diff.fetch(8, size);
                rslt_diff.fetch(10, md5sum);
                std::cout << mtime << "  " << std::setw(8) << size << "  "
                          << md5sum << "  ";
              }
              std::cout << xpath << lname << std::endl
                        << "+f  ";
              if (verbose) {
                std::string mtime, size, md5sum;
                rslt_diff.fetch(7, mtime);
                rslt_diff.fetch(9, size);
                rslt_diff.fetch(11, md5sum);
                std::cout << mtime << "  " << std::setw(8) << size << "  "
                          << md5sum << "  ";
              }
              std::cout << xpath << lname << std::endl;
            }
            else if (ltype == 1 && rtype == 0) {
              std::cout << "-f  ";
              if (verbose) {
                std::string mtime, size, md5sum;
                rslt_diff.fetch(6, mtime);
                rslt_diff.fetch(8, size);
                rslt_diff.fetch(10, md5sum);
                std::cout << mtime << "  " << std::setw(8) << size << "  "
                          << md5sum << "  ";
              }
              std::cout << xpath << lname << std::endl
                        << "+d  ";
              if (verbose) {
                std::string mtime, size, md5sum;
                rslt_diff.fetch(7, mtime);
                rslt_diff.fetch(9, size);
                rslt_diff.fetch(11, md5sum);
                std::cout << mtime << "  " << std::setw(8) << size << "  "
                          << md5sum << "  ";
              }
              std::cout << xpath << lname << std::endl;
            }
            else
              throw myexception("Unknown type in database entry.");
          }
        }
      };

      std::string path;
      do_diff(0, 0, "");
    }



    if (cmd_str == "expire") {
      std::vector<sqlid_t> ids_expired;
      cppdb::statement s = sql <<
        "SELECT id, created, basedir FROM toc"
        "  WHERE id <> ("
        "    SELECT MAX(id) FROM toc WHERE basedir = ?"
        "  )"
        "  AND basedir = ?;";
      s.bind(basedir);
      s.bind(basedir);
      cppdb::result r = s.query();

      while (r.next()) {
        sqlid_t id = 0;
        std::string created, name;
        r >> id >> created >> name;
        std::cerr << "removing " << name << " @ " << id
                  << " (" << created << ")" << std::endl;
        ids_expired.push_back(id);
      }

      // reopen database to prevent deadlocks
      sql.close();
      sql.open("sqlite3:db=" + db_name);

      cppdb::transaction guard(sql);
      for (const auto id: ids_expired) {
        sql << "DROP INDEX IF EXISTS idx_id_parent_tree" + std::to_string(id)
               + ";" << cppdb::exec;
        sql << "DROP INDEX IF EXISTS idx_inode_tree"     + std::to_string(id)
               + ";" << cppdb::exec;
        sql << "DROP TABLE tree"   + std::to_string(id) + ";" << cppdb::exec;
        sql << "DROP TABLE inodes" + std::to_string(id) + ";" << cppdb::exec;
        sql << "DELETE FROM toc WHERE id = " + std::to_string(id) + ";"
            << cppdb::exec;
      }
      guard.commit();

      sql << "VACUUM;" << cppdb::exec;
    }



    if (cmd_str == "stats") {
      std::cout << "filesystem updates:" << std::endl;
      cppdb::result r5 = sql <<
        "SELECT id, created, basedir FROM toc;";
      while (r5.next()) {
        std::string id, created, basedir;
        r5 >> id >> created >> basedir;
        std::cout << id << ": " << basedir
                  << " (" << created << ")" << std::endl;
      }

      std::cout << std::endl
                << "statistics (update " << id_last << "):" << std::endl;
      cppdb::result r0 = sql <<
        "SELECT COUNT(id) FROM tree" + id_last + " t"
        "  JOIN inodes" + id_last + " i"
        "  ON t.inode = i.inode"
        "  WHERE i.type = 0;" << cppdb::row;
      if (! r0.empty()) {
        sqlid_t n;
        r0 >> n;
        std::cout << "number of directories: "
                  << n << std::endl;
      }

      cppdb::result r1 = sql <<
        "SELECT COUNT(id) FROM tree" + id_last + " t"
        "  JOIN inodes" + id_last + " i"
        "  ON t.inode = i.inode"
        "  WHERE i.type = 1;" << cppdb::row;
      if (! r1.empty()) {
        sqlid_t n;
        r1 >> n;
        std::cout << "number of files:       "
                  << n << std::endl;
      }

      cppdb::result r3 = sql <<
        "SELECT COUNT(inode) FROM inodes" + id_last +
        "  WHERE type = 1;" << cppdb::row;
      if (! r3.empty()) {
        sqlid_t n;
        r3 >> n;
        std::cout << "number of inodes:      "
                  << n << std::endl;
      }

      cppdb::result r2 = sql <<
        "SELECT size FROM tree" + id_last + " t"
        "  JOIN inodes" + id_last + " i"
        "  ON t.inode = i.inode"
        "  WHERE t.id_parent IS NULL;" << cppdb::row;
      if (! r2.empty()) {
        off_t size;
        r2 >> size;
        std::cout << "size  [bytes]:         "
                  << size << std::endl;
      }

      cppdb::result r4 = sql <<
        "SELECT SUM(size) FROM inodes" + id_last +
        "  WHERE type = 1;" << cppdb::row;
      if (! r4.empty()) {
        off_t size;
        r4 >> size;
        std::cout << "usage [bytes]:         "
                  << size << std::endl;
      }
    }



    return retval;
  }
  catch(std::exception const &e) {
    std::cerr << "ERROR: " << e.what() << std::endl;
    return 1;
  }
  return 0;
}
